from base64 import b64encode
from os import path

from ajprax.files import contents, crc32c, md5, sha1, sha256
from tests import iter_eq


def asset(name):
    return path.join(path.dirname(__file__), "assets", name)


def test_contents():
    def test(filename, chunks, *a, **kw):
        assert iter_eq(contents(filename, *a, **kw), chunks)

    test(asset("empty"), [])
    test(asset("empty"), [], binary=True)
    test(asset("empty"), [], chunk_size=1)
    test(asset("empty"), [], binary=True, chunk_size=5)
    test(asset("stuff"), ["some text or something"])
    test(asset("stuff"), [b"some text or something"], binary=True)
    test(asset("stuff"), ["some ", "text ", "or so", "methi", "ng"], chunk_size=5)
    test(asset("stuff"), [b"some ", b"text ", b"or so", b"methi", b"ng"], binary=True, chunk_size=5)


def test_hashers():
    def test(hash, empty, stuff):
        def test(filename, raw):
            assert hash(asset(filename)) == raw
            assert hash(asset(filename), b64encode) == b64encode(raw)

        test("empty", empty)
        test("stuff", stuff)

    test(
        crc32c,
        b"\x00\x00\x00\x00",
        b"\xf7\x88\xc4\x84",
    )
    test(
        md5,
        b"\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~",
        b"P\xd22\x8b\x9a\x85!\xc2\xefm\xc8\xde\x06\xba\x12\xa9"
    )
    test(
        sha1,
        b"\xda9\xa3\xee^kK\r2U\xbf\xef\x95`\x18\x90\xaf\xd8\x07\t",
        b"\xcaFe\x8a\x95\x192\xa2\xcbu\x89\xed\x7fs\xbf|\x86\xb3\x14\x80"
    )
    test(
        sha256,
        b"\xe3\xb0\xc4B\x98\xfc\x1c\x14\x9a\xfb\xf4\xc8\x99o\xb9$'\xaeA\xe4d\x9b\x93L\xa4\x95\x99\x1bxR\xb8U",
        b"o\x11z\xfe\xd94\xc27{5\xdf\x98q\xa6\xfc\xb1\x95\xcc\xa27#\xe7\xca\xd4\xbe\xf2\x0f6\x1e\xd7K\xed",
    )



def identity(x):
    return x


def kw(f):
    """
    Convert a function to take a single dictionary of keyword arguments

    Similar to, but more explicit that `t`
    Example:
         def add(a=0, b=0):
             return a + b

        # imagine that the standard library had nice collections
        kwargs = [dict(a=5), dict(b=10), dict(a=5, b=10)]
        kwargs.map(lambda d: add(d.get("a", 0), d.get("b", 0)))
        # no manual decomposition of the dict, use functions that weren't written for use with HOFs
        kwargs.map(kw(add))
        # use more readable lambdas
        kwargs.map(kw(lambda a=0, b=0: a + b))
    """

    def kw(kw):
        return f(**kw)

    return kw


def t(f):
    """
    Convert a function of N arguments to a function of one N-degree tuple

    Useful with higher order functions that expect a single-argument function as input. This is given as an alternative
    to implementing a star_map, star_for_each, star_flat_map, etc functions for all collections.

    Example:
        def add(a, b):
            return a + b

        # imagine that the standard library had nice collections
        pairs = [(1, 2), (2, 4), (3, 6)]
        pairs.map(lambda pair: add(pair[0], pair[1]))
        # no manual decomposition of the pair, use functions that weren't written for use with HOFs
        pairs.map(t(add))
        # use more readable lambdas
        pairs.map(t(lambda a, b: a + b))
    """

    def t(t):
        return f(*t)

    return t



import builtins
import os


def print(*values, sep=" ", end=os.linesep, file=None, flush=False):
    """Only outwardly visible behavior change from builtins.print is that the content is printed atomically."""
    builtins.print(sep.join(str(value) for value in values) + end, end="", file=file, flush=flush)



from datetime import datetime, timezone

from ajprax.collections import Iter
from ajprax.logging import LEVEL_NAME, Logger, Log, INFO


class TestLog:
    def test_level(self):
        def test(log_level, message_level, *a, **kw):
            def check(log):
                # log level is passed through to the Log object
                assert log.level == message_level
                # log is only sent if the message level meets or exceeds the log level
                assert message_level >= log_level

            logger = Logger()
            logger.level = log_level
            logger.subscribe(check)
            name = LEVEL_NAME[message_level].strip().lower()
            getattr(logger, name)(*a, **kw)

        for log_level, message_level in Iter(LEVEL_NAME).product(repeat=2):
            test(log_level, message_level)
            test(log_level, message_level, "message")
            test(log_level, message_level, key="value")
            test(log_level, message_level, "message", key="value")

    def test_format(self):
        def test(datetime, level, message, keywords, exception):
            log = str(Log(datetime, level, message, keywords, exception))
            assert ("log message" in log) == bool(message)
            assert ("key='value'" in log) == bool(keywords)
            assert ("exception message" in log) == bool(exception)
            assert log.split("\n")[0].startswith("2024-01-01T00:00:00.000001Z  INFO ")

        for (message, keywords, exception) in Iter((True, False)).product(repeat=3):
            message = "log message" if message else ""
            keywords = {"key": "value"} if keywords else {}
            exception = Exception("exception message") if exception else None
            test(
                datetime(2024, 1, 1, 0, 0, 0, 1, tzinfo=timezone.utc),
                INFO,
                message,
                keywords,
                exception,
            )



from ajprax.hof import identity
from ajprax.sentinel import Unset
from abc import abstractmethod, ABC
from queue import Queue

from ajprax.collections import Iter


class Notifications:
    @classmethod
    def on(cls, on):
        notifications = Notifications()
        on.subscribe(lambda *a, **kw: notifications.notify())
        return notifications

    def __init__(self):
        self._callbacks = []

    def subscribe(self, callback):
        self._callbacks.append(callback)
        return Unsubscribe(self, callback)

    def unsubscribe(self, callback):
        self._callbacks.remove(callback)

    def notify(self, *a, **kw):
        for callback in self._callbacks:
            callback(*a, **kw)


class Unsubscribe:
    def __init__(self, notifications, callback):
        self.notifications = notifications
        self.callback = callback

    def __call__(self, *args, **kwargs):
        self.notifications.unsubscribe(self.callback)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self()


class Events(Notifications):
    def send(self, event):
        self.notify(event)

    def __iter__(self):
        return IterEvents(self)

    def all(self, key=Unset):
        pass

    def any(self, key=Unset):
        pass

    def batch(self, size):
        pass

    def count(self, key=Unset):
        pass

    def distinct(self, key=Unset):
        pass

    def filter(self, fn):
        return FilteredEvents(self, fn)

    def flat_map(self, fn):
        return FlatMappedEvents(self, fn)

    def map(self, fn):
        return MappedEvents(self, fn)

    def on(self):
        return Notifications.on(self)


class Value(Notifications):
    def __iadd__(self, other):
        self.value += other
        return self

    def __iand__(self, other):
        self.value &= other
        return self

    def __idiv__(self, other):
        self.value /= other
        return self

    def __ifloordiv__(self, other):
        self.value //= other
        return self

    def __ilshift__(self, other):
        self.value <<= other
        return self

    def __imatmul__(self, other):
        self.value @= other
        return self

    def __imod__(self, other):
        self.value %= other
        return self

    def __init__(self, initial=Unset):
        Notifications.__init__(self)
        self._value = initial

    def __imul__(self, other):
        self.value *= other
        return self

    def __ior__(self, other):
        self.value |= other
        return self

    def __ipow__(self, other):
        self.value **= other
        return self

    def __irshift__(self, other):
        self.value >>= other
        return self

    def __isub__(self, other):
        self.value -= other
        return self

    def __itruediv__(self, other):
        self.value /= other
        return self

    def __ixor__(self, other):
        self.value ^= other
        return self

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self.set(value)

    def set(self, value):
        self._value = value
        self.notify(value)

    def subscribe(self, callback):
        if self._value is not Unset:
            callback(self._value)
        return Notifications.subscribe(self, callback)

    def all(self, key=Unset):
        return AllValue(self, key)

    def any(self, key=Unset):
        return AnyValue(self, key)

    def changed(self):
        return self.changes().on()

    def changes(self):
        return Changes(self)

    def has_been(self, predicate=bool):
        return HasBeen(self, predicate)

    def is_(self, predicate=bool):
        return Is(self, predicate)

    def map(self, fn):
        return MappedValue(self, fn)

    def on(self):
        return Notifications.on(self)

    def zip(self, *others):
        return ZippedValue(self, *others)


class EventsCombinator(ABC, Events):
    def __init__(self, events):
        Events.__init__(self)
        self.events = events
        self.events.subscribe(self.on_event)

    @abstractmethod
    def on_event(self, event):
        pass

    def close(self):
        self.events.unsubscribe(self.on_event)


class AllEvents(Value):
    def __init__(self, events, key=Unset):
        self.key = identity if key is Unset else key
        Value.__init__(self)
        self.set(True)
        events.subscribe(self.on_event)

    def on_event(self, event):
        self.value &= bool(self.key(event))


class IterEvents(EventsCombinator, Iter):
    def __init__(self, events):
        def iter_from_queue():
            while True:
                yield self.queue.get()

        self.queue = Queue()
        Iter.__init__(self, iter_from_queue())
        EventsCombinator.__init__(self, events)

    def on_event(self, event):
        print("putting", event)
        self.queue.put(event)


class FilteredEvents(EventsCombinator):
    def __init__(self, events, fn):
        self.fn = fn
        EventsCombinator.__init__(self, events)

    def on_event(self, event):
        if self.fn(event):
            self.send(event)


class FlatMappedEvents(EventsCombinator):
    def __init__(self, events, fn):
        self.fn = fn
        EventsCombinator.__init__(self, events)

    def on_event(self, event):
        for event in self.fn(event):
            self.send(event)


class MappedEvents(EventsCombinator):
    def __init__(self, events, fn):
        self.fn = fn
        EventsCombinator.__init__(self, events)

    def on_event(self, event):
        self.send(self.fn(event))


class ValueCombinator(ABC, Value):
    def __init__(self, value):
        Value.__init__(self)
        self.v = value
        self.v.subscribe(self.on_set)

    @abstractmethod
    def on_set(self, value):
        pass

    def close(self):
        self.v.unsubscribe(self.on_set)


class AllValue(ValueCombinator):
    def __init__(self, value, key):
        self.key = identity if key is Unset else key
        ValueCombinator.__init__(self, value)
        self.set(True)

    def on_set(self, value):
        self.value &= bool(self.key(value))


class AnyValue(ValueCombinator):
    def __init__(self, value, key):
        self.key = identity if key is Unset else key
        ValueCombinator.__init__(self, value)
        self.set(False)

    def on_set(self, value):
        self.value |= bool(self.key(value))


class Changes(ValueCombinator):
    def on_set(self, value):
        if value != self.value:
            self.set(value)


class HasBeen(ValueCombinator):
    def __init__(self, value, predicate):
        self.predicate = predicate
        ValueCombinator.__init__(self, value)

    def on_set(self, value):
        if self.value is Unset:
            self.value = self.predicate(value)
        else:
            self.value |= self.predicate(value)


class Is(ValueCombinator):
    def __init__(self, value, predicate):
        self.predicate = predicate
        ValueCombinator.__init__(self, value)

    def on_set(self, value):
        self.set(self.predicate(value))


class MappedValue(ValueCombinator):
    def __init__(self, value, fn):
        self.fn = fn
        ValueCombinator.__init__(self, value)

    def on_set(self, value):
        self.set(self.fn(value))


class ZippedValue(Value):
    def __init__(self, *values):
        Value.__init__(self)
        raw = [Unset] * len(values)
        for i, value in enumerate(values):
            # i=i is here because if we allow seti to close over i then all copies will see the highest value of i
            # instead of their particular value
            def seti(value, i=i):
                raw[i] = value
                # TODO: avoid checking this every time
                if Unset not in raw:
                    self.set(tuple(raw))

            value.subscribe(seti)
